import logging
import aiohttp.web

from typing import Optional
from aiohttp.web import Application, TCPSite, AppRunner, Request
from discord.ext.ipc.errors import *
from discord.ext.commands import Bot, Cog

log = logging.getLogger(__name__)

def route(name: Optional[str] = None):
    """
    |method|
    
    Used to register a coroutine as an endpoint when you don't have
    access to an instance of :class:`.Server`
    Parameters
    ----------
    name: :str:`str`
        The endpoint name. If not provided the method name will be
        used.
    """
    def decorator(func):
        if not name:
            Server.endpoints[func.__name__] = func
        else:
            Server.endpoints[name] = func
        return func
    return decorator

class IpcServerResponse:
    def __init__(self, data):
        self._json = data
        self.length = len(data)

        self.endpoint = data["endpoint"]

        for key, value in data["data"].items():
            setattr(self, key, value)

    def to_json(self):
        return self._json

    def __repr__(self):
        return "<IpcServerResponse length={0.length}>".format(self)

    def __str__(self):
        return self.__repr__()

class Server:
    """ 
    |class|
    The IPC server. Usually used on the bot process for receiving
    requests from the client.
    Attributes
    ----------
    bot: :class:`~discord.ext.commands.Bot`
        Your bot instance
    host: :str:`str`
        The host to run the IPC Server on. Defaults to `127.0.0.1`.
    port: :str:`int`
        The port to run the IPC Server on. Defaults to 1025.
    secret_key: :str:`str`
        A secret key. Used for authentication and should be the same as
        your client's secret key.
    do_multicast: :bool:`bool`
        Turn multicasting on/off. Defaults to False
    multicast_port: :int:`int`
        The port to run the multicasting server on. Defaults to 20000
    logger: `logging.Logger`
        A custom logger for all event. Default on is `discord.ext.ipc`
    """

    def __init__(
        self, 
        bot: Bot, 
        host: str = "127.0.0.1", 
        port: int = 1025,
        secret_key: str = None, 
        do_multicast: bool = False,
        multicast_port: int = 20000,
        logger: logging.Logger = log
    ):
        self.bot = bot
        self.host = host
        self.port = port
        self.secret_key = secret_key
        self.do_multicast = do_multicast
        self.multicast_port = multicast_port
        self.logger = logger

    endpoints = {}
    _server = None
    _multicast_server = None
    _cls = None

    def start(self, cls: Cog) -> None:
        """
        |method|
        
        Starts the IPC server
        Parameters
        ----------
        cls: `~discord.ext.commands.Cog`
            The Cog where all the routes are located.
        """
        self._server = Application()
        self._server.router.add_route("GET", "/", self.handle_accept)
        self._cls = cls.__cog_name__

        if self.do_multicast:
            self._multicast_server = Application()
            self._multicast_server.router.add_route("GET", "/", self.handle_multicast)
            self.bot.loop.create_task(self.setup(self._multicast_server, self.multicast_port))

        self.bot.loop.create_task(self.setup(self._server, self.port))
        self.bot.dispatch("ipc_ready")
        self.logger.info("The IPC server is ready")

    def route(self, name: Optional[str] = None):
        """
        |method|

        Used to register a coroutine as an endpoint when you have
        access to an instance of :class:`~discord.ext.ipc.Server`
        
        Please note that the endpoints are registered in :function:`update_endpoints`

        Parameters
        ----------
        name: `str`
            The endpoint name. If not provided the method name will be used.
        """

        def decorator(func):
            if not name:
                self.endpoints[func.__name__] = func
            else:
                self.endpoints[name] = func
            return func
        return decorator

    async def handle_accept(self, request: Request) -> None:
        """
        |coro|

        Handles websocket requests from the client process.

        Parameters
        ----------
        request: :class:`~aiohttp.web.Request`
            The request made by the client, parsed by aiohttp.
        """
        self.logger.info("Handing new IPC request")

        websocket = aiohttp.web.WebSocketResponse()
        await websocket.prepare(request)

        async for message in websocket:
            request = message.json()

            self.logger.debug("IPC Server < %r", request)

            endpoint = request.get("endpoint")
            headers = request.get("headers")

            if not headers or headers.get("Authorization") != self.secret_key:
                self.bot.dispatch("ipc_error", endpoint, Exception("Received unauthorized request (Invalid or no token provided)."))
                response = {
                    "error": "Received unauthorized request (invalid or no token provided).", 
                    "code": 403
                }
            else:
                if not endpoint or endpoint not in self.endpoints:
                    self.bot.dispatch("ipc_error", endpoint, Exception("Received invalid request (invalid or no endpoint given)."))
                    response = {
                        "error": "Received invalid request (invalid or no endpoint given).",
                        "code": 400
                    }
                else:
                    server_response = IpcServerResponse(request)

                    try:
                        if (attempted_cls := self.bot.cogs.get(self.endpoints[endpoint].__qualname__.split(".")[0])):
                            arguments = (attempted_cls, server_response)
                        else:
                            guaranteed_cls = self.bot.cogs.get(self._cls)
                            arguments = (guaranteed_cls, server_response)
                    except AttributeError:
                        arguments = (server_response, )

                    self.logger.debug(arguments)

                    try:
                        response = await self.endpoints[endpoint](*arguments)
                    except Exception as error:
                        self.logger.error(
                            "Received error while executing %r with %r",
                            endpoint,
                            request,
                            exc_info=error
                        )
                        self.bot.dispatch("ipc_error", endpoint, error)

                        response = {
                            "error": str(error),
                            "code": 500,
                        }

            try:
                if not response: response = {}
                if not response.get("code"):
                    response["code"] = 200

                await websocket.send_json(response)
                self.logger.debug("IPC Server > %r", response)
            except TypeError as error:
                if str(error).startswith("Object of type") and str(error).endswith("is not JSON serializable"):
                    error_response = (
                        "IPC route returned values which are not able to be sent over sockets."
                        "If you are trying to send a discord.py object,"
                        "please only send the data you need."
                    )

                    self.bot.dispatch("ipc_error", endpoint, Exception(error_response))

                    response = {
                        "error": error_response, 
                        "code": 500
                    }

                    await websocket.send_json(response)
                    self.logger.debug("IPC Server > %r", response)

                    raise JSONEncodeError(error_response)

    async def handle_multicast(self, request: Request) -> None:
        """
        |coro|
        Handles multicasting websocket requests from the client.
        Parameters
        ----------
        request: :class:`~aiohttp.web.Request`
            The request made by the client, parsed by aiohttp.
        """
        self.logger.info("Initiating Multicast Server.")
        websocket = aiohttp.web.WebSocketResponse()
        await websocket.prepare(request)

        async for message in websocket:
            request = message.json()

            log.debug("Multicast Server < %r", request)

            headers = request.get("headers")

            if not headers or headers.get("Authorization") != self.secret_key:
                response = {"error": "Invalid or no token provided.", "code": 403}
            else:
                response = {
                    "message": "Connection success",
                    "port": self.port,
                    "code": 200,
                }

            self.logger.debug("Multicast Server > %r", response)
            await websocket.send_json(response)

    async def setup(self, application: Application, port: int) -> None:
        """
        |coro|
        This function stats the IPC runner and the IPC webserver
        Parameters
        ----------
        application: :class:`aiohttp.web.Application`
            The internal router's app
        port: :int:`int`
            The specific port to run the application (:class:`~aiohttp.web.Application`)
        """
        self.logger.debug('Starting the IPC runner')
        self._runner = AppRunner(application)
        await self._runner.setup()

        self.logger.debug('Starting the IPC webserver')
        self._webserver = _webserver = TCPSite(self._runner, self.host, port)
        await _webserver.start()

    async def stop(self) -> None:
        """
        |coro|
        Stops both the IPC webserver
        """
        self.logger.info('Stopping up the IPC webserver')
        self.logger.debug(self._runner.addresses)
        await self._webserver.stop()



from __future__ import annotations
import asyncio
import logging
import typing
import aiohttp

from typing import Optional, Union, Any
from aiohttp import ClientWebSocketResponse
from discord.ext.ipc.errors import *

log = logging.getLogger(__name__)

class Client:
    """
    |class|
    
    Handles webserver side requests to the bot process.

    Parameters
    ----------
    host: str
        The IP or host of the IPC server, defaults to `127.0.0.1`
    port: int
        The port of the IPC server. If not supplied the port will be found automatically, defaults to None
    secret_key: Union[str, bytes]
        The secret key for your IPC server. Must match the server secret_key or requests will not go ahead, defaults to None
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: Optional[int] = None,
        multicast_port: int = 20000,
        secret_key: Union[str, bytes] = None
    ):
        """Constructor"""
        self.host = host
        self.port = port
        self.secret_key = secret_key
        self.multicast_port = multicast_port
        self.lock = None
        self.loop = None
        self.logger = None
        self.session = None
        self.websocket = None
        self.multicast = None
        self.closed = False

    @property
    def url(self) -> str:
        return f"ws://{self.host}:{self.port if self.port else self.multicast_port}"

    async def init_sock(self) -> ClientWebSocketResponse:
        """
        Attempts to connect to the server

        Returns
        -------
        :class: `~aiohttp.ClientWebSocketResponse`
            The websocket connection to the server
        """
        self.logger.info("Initiating WebSocket connection.")
        self.session = aiohttp.ClientSession()

        if not self.port:
            self.logger.debug("No port was provided - initiating multicast connection at %s.", self.url,)
            self.multicast = await self.session.ws_connect(self.url, autoping=False)

            payload = {"connect": True, "headers": {"Authorization": self.secret_key}}
            self.logger.debug("Multicast Server < %r", payload)

            await self.multicast.send_json(payload)
            recv = await self.multicast.receive()

            self.logger.debug("Multicast Server > %r", recv)

            if recv.type in (aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSED):
                self.logger.error(
                    "WebSocket connection unexpectedly closed. Multicast Server is unreachable."
                )
                raise NotConnected("Multicast server connection failed.")

            port_data = recv.json()
            self.port = port_data["port"]

        self.websocket = await self.session.ws_connect(self.url, autoping=False, autoclose=False)
        self.logger.info("Client connected to %s", self.url)

        return self.websocket

    async def request(
        self,
        endpoint: str, **kwargs
    ) -> Optional[Any]:
        """
        |coro|

        Make a request to the IPC server process.

        Parameters
        ----------
        endpoint: `str`
            The endpoint to request on the server
        **kwargs
            The data to send to the endpoint
        """
        if not self.session:
            raise RuntimeError("The IPC has not been started yet!")

        if self.closed:
            raise RuntimeError("The IPC is currently closed!")

        self.logger.info("Requesting IPC Server for %r with %r", endpoint, kwargs)
        
        payload = {
            "endpoint": endpoint,
            "data": kwargs,
            "headers": {"Authorization": self.secret_key},
        }

        try:
            await self.websocket.send_json(payload)
        except ConnectionResetError:
            self.logger.error("Cannot write to closing transport, attempting reconnection in 5 seconds.")
            await self.session.close()
            await asyncio.sleep(5)
            await self.init_sock()

            return await self.request(endpoint, **kwargs)

        self.logger.debug("Client > %r", payload)

        async with self.lock:
            recv = await self.websocket.receive()

        self.logger.debug("Client < %r", recv)

        if recv.type == aiohttp.WSMsgType.PING:
            self.logger.info("Received request to PING")
            await self.websocket.ping()
            return await self.request(endpoint, **kwargs)

        elif recv.type == aiohttp.WSMsgType.PONG:
            self.logger.info("Received PONG")
            return await self.request(endpoint, **kwargs)

        elif recv.type == aiohttp.WSMsgType.CLOSED:
            self.logger.error("WebSocket connection unexpectedly closed, attempting reconnection in 5 seconds.")
            await self.session.close()
            await asyncio.sleep(5)
            await self.init_sock()

            return await self.request(endpoint, **kwargs)
        
        else:
            return recv.json()

    async def start(
        self, 
        loop: Optional[asyncio.AbstractEventLoop] = None,
        logger: Optional[logging.Logger] = None
    ) -> Client:
        """
        |coro|

        Starts the IPC session

        Parameters
        ----------
        loop: `asyncio.AbstractEventLoop`
            Asyncio loop to create the server in, if None, take default one.
            If specified it is the caller's responsibility to close and cleanup the loop.
        logger: `logging.Logger`
            A custom logger for all event. Default on is `discord.ext.ipc`
        """
        if not loop:
            loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        if not logger:
            logger = log
        
        self.loop = loop
        self.logger = logger
        self.lock = asyncio.Lock()
        await self.init_sock()

        return self
    
    async def close(self) -> None:
        """
        |coro|

        Stops the IPC session
        """
        if self.session:
            await self.session.close()
        self.closed = True
    



"""
Better IPC
~~~~~~
A high-performance inter-process communication 
library designed to work with the latest version of discord.py
:license: Apache License 2.0
"""

try:
    import discord
except ImportError:
    raise RuntimeError("You must have discord.py installed!")
else:
    if not discord.__version__.startswith("2.0"):
        raise RuntimeError(
            "You must have discord.py 2.0 installed in order for Better IPC to work!"
        )


__version__ = "1.0.3"
__title__ = "better-ipc"
__author__ = "DaPandaOfficial"

from discord.ext.ipc.client import Client
from discord.ext.ipc.server import Server
from discord.ext.ipc.errors import *



import setuptools
from pathlib import Path

with open("requirements.txt") as stream:
    raw = stream.read().splitlines()
    requirements = [x for x in raw if not x.startswith("git+")]
    dependencies = [x for x in raw if x.startswith("git+")]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(
    author="DaPandaOfficial",
    python_requires=">=3.8.0",
    license="Apache Software License",
    author_email="miroslav.rosenov39@gmail.com",
    long_description_content_type="text/markdown",
    url="https://github.com/MiroslavRosenov/better-ipc",
    download_url="https://github.com/MiroslavRosenov/better-ipc/archive/refs/tags/1.0.tar.gz",
    description="A high-performance inter-process communication library designed to work with the latest version of discord.py",
    packages=[
        "discord.ext.ipc"
    ],
    project_urls={
        "Source": "https://github.com/MiroslavRosenov/better-ipc",
        "Issue Tracker": "https://github.com/MiroslavRosenov/better-ipc/issues",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["better_ipc", "ipc", "python", "discord.py"],
    long_description=long_description,
    install_requires=requirements,
    dependencies=dependencies,
    name="better-ipc",
    version="1.0.3",
)



from discord import ClientException
from typing import Tuple

class IPCError(ClientException):
    """Base IPC exception class"""
    __slots__: Tuple[str, ...] = ()


class NoEndpointFoundError(IPCError):
    """Raised upon requesting an invalid endpoint"""
    pass


class ServerConnectionRefusedError(IPCError):
    """Raised upon a server refusing to connect / not being found"""
    pass


class JSONEncodeError(IPCError):
    """Raise upon un-serializable objects are given to the IPC"""
    pass


class NotConnected(IPCError):
    """Raised upon websocket not connected"""
    pass
